==> azureml_train_automl-1.0.6-py3-none-any.whl: azureml/train/automl/_win32_helper.py <==
# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""_win32_helper.py, Helper code that wraps the Win32 API using ctypes"""

import ctypes
from ctypes import wintypes
from ctypes import windll


class PROCESS_MEMORY_COUNTERS_EX(ctypes.Structure):
    _fields_ = [('cb', wintypes.DWORD),
                ('PageFaultCount', wintypes.DWORD),
                ('PeakWorkingSetSize', wintypes.WPARAM),
                ('WorkingSetSize', wintypes.WPARAM),
                ('QuotaPeakPagedPoolUsage', wintypes.WPARAM),
                ('QuotaPagedPoolUsage', wintypes.WPARAM),
                ('QuotaPeakNonPagedPoolUsage', wintypes.WPARAM),
                ('QuotaNonPagedPoolUsage', wintypes.WPARAM),
                ('PagefileUsage', wintypes.WPARAM),
                ('PeakPagefileUsage', wintypes.WPARAM),
                ('PrivateUsage', wintypes.WPARAM)]


class IO_COUNTERS(ctypes.Structure):
    _fields_ = [('ReadOperationCount', wintypes.ULARGE_INTEGER),
                ('WriteOperationCount', wintypes.ULARGE_INTEGER),
                ('OtherOperationCount', wintypes.ULARGE_INTEGER),
                ('ReadTransferCount', wintypes.ULARGE_INTEGER),
                ('WriteTransferCount', wintypes.ULARGE_INTEGER),
                ('OtherTransferCount', wintypes.ULARGE_INTEGER)]


class JOBOBJECT_BASIC_ACCOUNTING_INFORMATION(ctypes.Structure):
    _fields_ = [('TotalUserTime', wintypes.LARGE_INTEGER),
                ('TotalKernelTime', wintypes.LARGE_INTEGER),
                ('ThisPeriodTotalUserTime', wintypes.LARGE_INTEGER),
                ('ThisPeriodTotalKernelTime', wintypes.LARGE_INTEGER),
                ('TotalPageFaultCount', wintypes.DWORD),
                ('TotalProcesses', wintypes.DWORD),
                ('ActiveProcesses', wintypes.DWORD),
                ('TotalTerminatedProcesses', wintypes.DWORD)]


class JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION(ctypes.Structure):
    _fields_ = [('BasicInfo', JOBOBJECT_BASIC_ACCOUNTING_INFORMATION),
                ('IoInfo', IO_COUNTERS)]


class JobObjectInfoClass(object):
    JobObjectBasicAccountingInformation = 1
    JobObjectBasicAndIoAccountingInformation = 8
    JobObjectBasicLimitInformation = 2
    JobObjectBasicProcessIdList = 3
    JobObjectBasicUIRestrictions = 4
    JobObjectCpuRateControlInformation = 15
    JobObjectEndOfJobTimeInformation = 6
    JobObjectExtendedLimitInformation = 9
    JobObjectGroupInformation = 11
    JobObjectGroupInformationEx = 14
    JobObjectLimitViolationInformation = 13
    JobObjectLimitViolationInformation2 = 35
    JobObjectNetRateControlInformation = 32
    JobObjectNotificationLimitInformation = 12
    JobObjectNotificationLimitInformation2 = 34


class Win32Helper(object):

    def __init__(self):
        # Get a handle to ourselves for future use

        # HANDLE GetCurrentProcess()
        self.current_process = windll.kernel32.GetCurrentProcess()

        self.child_handles = {}

        if not self.is_process_in_job():
            # Create a job with the name automl-{PID} and assign ourselves to it. All child processes will inherit
            # this job object so we can look them up later.
            # We don't need to hold a reference to the job object because passing NULL instead of a job handle
            # will make the Win32 API use this process's job.

            # HANDLE CreateJobObjectW(LPSECURITY_ATTRIBUTES lpJobAttributes, LPWSTR lpName)
            job_object = windll.kernel32.CreateJobObjectW(None, 'automl-{}'.format(self.get_process_id()))

            self.assign_process_to_job_object(job_object)

    def __del__(self):
        # Clean up all handles to child processes to prevent memory leaks
        for pid in self.child_handles:
            self.close_handle(self.child_handles[pid])

    def close_handle(self, handle):
        """
        Wrapper for the Win32 CloseHandle function.
        BOOL WINAPI CloseHandle(
          _In_ HANDLE hObject
        );

        :param handle: A valid handle to an open object.
        """
        success = windll.kernel32.CloseHandle(handle)
        if success == 0:
            raise self.get_error('CloseHandle')

    def get_process_id(self):
        """
        Wrapper for Win32 GetProcessId function.
        DWORD GetProcessId(
          HANDLE Process
        );

        :return: the pid of the current process
        """
        return windll.kernel32.GetProcessId(self.current_process)

    def open_process(self, pid=None):
        """
        Wrapper for the Win32 OpenProcess function.
        Note: If a handle was already opened, the previously opened handle is returned instead. This is to prevent
        memory leaks from opening too many handles and then failing to close them. This way, we only need to close
        handles once (when this object is destroyed).
        HANDLE OpenProcess(
          DWORD dwDesiredAccess,
          BOOL  bInheritHandle,
          DWORD dwProcessId
        );

        :param pid: The PID of the process to create a handle for. If None, uses the current process.
        :return: the handle of the process
        """
        if pid is None:
            return self.current_process

        if pid in self.child_handles:
            return self.child_handles[pid]

        PROCESS_ALL_ACCESS = (0x000F0000 | 0x00100000 | 0xFFF)
        handle = windll.kernel32.OpenProcess(PROCESS_ALL_ACCESS, False, pid)
        if handle == 0:
            raise self.get_error('OpenProcess')
        self.child_handles[pid] = handle
        return handle

    def get_process_memory_info(self, pid=None):
        """
        Wrapper for the Win32 GetProcessMemoryInfo function.
        BOOL GetProcessMemoryInfo(
          HANDLE                   Process,
          PPROCESS_MEMORY_COUNTERS ppsmemCounters,
          DWORD                    cb
        );

        :param pid: ID of the process we want to get memory usage for. If None, gets usage for current process.
        :return: a PROCESS_MEMORY_COUNTERS_EX struct
        """
        handle = self.open_process(pid)
        pmc = PROCESS_MEMORY_COUNTERS_EX()
        success = windll.psapi.GetProcessMemoryInfo(handle, ctypes.byref(pmc), ctypes.sizeof(pmc))
        if success == 0:
            raise self.get_error('GetProcessMemoryInfo')
        return pmc

    def assign_process_to_job_object(self, job, pid=None):
        """
        Wrapper for the Win32 AssignProcessToJobObject function.
        BOOL WINAPI AssignProcessToJobObject(
          _In_ HANDLE hJob,
          _In_ HANDLE hProcess
        );
        :param job: Handle to the job object we want to assign to.
        :param pid: ID of the process we want to assign to the job. If None, assigns the current process to the job.
        """
        handle = self.open_process(pid)
        success = windll.kernel32.AssignProcessToJobObject(job, handle)
        if success == 0:
            raise self.get_error('AssignProcessToJobObject')

    def is_process_in_job(self):
        """
        Wrapper for the Win32 IsProcessInJob function.
        BOOL WINAPI IsProcessInJob(
          _In_     HANDLE ProcessHandle,
          _In_opt_ HANDLE JobHandle,
          _Out_    PBOOL  Result
        );

        :return: true if this process is running inside a job, false otherwise
        """
        result = ctypes.c_bool()
        success = windll.kernel32.IsProcessInJob(self.current_process, None, ctypes.byref(result))

        if success == 0:
            raise Win32Helper.get_error('IsProcessInJob')

        return result.value

    def query_information_job_object(self, job_object_info_class, job_info_struct):
        """
        Wrapper for the Win32 QueryInformationJobObject function. Queries info for the job this process is in.
        BOOL WINAPI QueryInformationJobObject(
          _In_opt_  HANDLE             hJob,
          _In_      JOBOBJECTINFOCLASS JobObjectInfoClass,
          _Out_     LPVOID             lpJobObjectInfo,
          _In_      DWORD              cbJobObjectInfoLength,
          _Out_opt_ LPDWORD            lpReturnLength
        );

        :param job_object_info_class: the info class that should be queried
        :param job_info_struct: a struct to pass to the query
        :return: a struct with the job info, type depends on which info class is specified
        """
        out_size = wintypes.DWORD()

        success = windll.kernel32.QueryInformationJobObject(None,
                                                            job_object_info_class,
                                                            ctypes.byref(job_info_struct),
                                                            ctypes.sizeof(job_info_struct),
                                                            ctypes.byref(out_size))
        if success == 0:
            raise self.get_error('QueryInformationJobObject')
        return job_info_struct

    def get_process_times(self, pid=None):
        """
        Wrapper for the Win32 GetProcessTimes function. Retrieves timing information for the specified process.
        Note: The Win32 API returns times in 100ns increments, but this function converts to seconds for convenience.
        BOOL GetProcessTimes(
          HANDLE     hProcess,
          LPFILETIME lpCreationTime,
          LPFILETIME lpExitTime,
          LPFILETIME lpKernelTime,
          LPFILETIME lpUserTime
        );

        :param pid: The PID of the process to get timing information for. If None, gets info for the current process.
        :return: a tuple of ints containing (creation time, exit time, kernel time, user time) in seconds
        """
        handle = self.open_process(pid)

        creation_time = wintypes.FILETIME()
        exit_time = wintypes.FILETIME()
        kernel_time = wintypes.FILETIME()
        user_time = wintypes.FILETIME()

        success = windll.kernel32.GetProcessTimes(handle,
                                                  ctypes.byref(creation_time),
                                                  ctypes.byref(exit_time),
                                                  ctypes.byref(kernel_time),
                                                  ctypes.byref(user_time))
        if success == 0:
            raise self.get_error('GetProcessTimes')

        def ft_to_sec(ft):
            total = ft.dwHighDateTime << 32
            total += ft.dwLowDateTime
            return total / 10000000

        return ft_to_sec(creation_time), ft_to_sec(exit_time), ft_to_sec(kernel_time), ft_to_sec(user_time)

    def get_current_memory_usage(self):
        """
        Gets the physical and virtual memory usage of this process.

        :return: a tuple containing [physical memory used, virtual memory used] in bytes
        """
        mem_info = self.get_process_memory_info()
        return mem_info.WorkingSetSize, mem_info.PrivateUsage

    def get_children_memory_usage(self):
        """
        Gets the physical and virtual memory usage of all of this process's children.

        :return: a tuple containing [physical memory used, virtual memory used] in bytes
        """
        pids = self.get_job_process_list()
        mem_info = [self.get_process_memory_info(pid) for pid in pids if pid != self.current_process]
        child_mem_usage = [0, 0]
        for info in mem_info:
            child_mem_usage[0] += info.WorkingSetSize
            child_mem_usage[1] += info.PrivateUsage
        return tuple(child_mem_usage)

    def get_job_accounting_info(self):
        """
        Gets the basic accounting and I/O accounting information for the job this process is in.

        :return: a JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION struct
        """
        struct = JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION()
        return self.query_information_job_object(JobObjectInfoClass.JobObjectBasicAndIoAccountingInformation, struct)

    def get_job_process_list(self):
        """
        Gets the list of pids of all processes associated with the job this process is in.

        :return: a list of all pids
        """
        captured_all_processes = False
        max_processes = 10
        struct = None
        while not captured_all_processes:
            # This must be defined dynamically because the struct is variable length
            class JobObjectBasicProcessIdListStruct(ctypes.Structure):
                _fields_ = [('NumberOfAssignedProcesses', wintypes.DWORD),
                            ('NumberOfProcessIdsInList', wintypes.DWORD),
                            ('ProcessIdList', wintypes.WPARAM * max_processes)]
            struct = JobObjectBasicProcessIdListStruct()
            struct.NumberOfAssignedProcesses = max_processes
            self.query_information_job_object(JobObjectInfoClass.JobObjectBasicProcessIdList, struct)
            if struct.NumberOfProcessIdsInList > struct.NumberOfAssignedProcesses:
                # Not enough space to capture all pids, resize array and try again
                max_processes *= 2
            else:
                captured_all_processes = True
        return list(struct.ProcessIdList[:struct.NumberOfProcessIdsInList])

    def get_child_process_times(self):
        """
        Gets the timing information for all currently running processes in this job, except this one.

        :return: a tuple of ints containing (creation time, exit time, kernel time, user time) in seconds
        """
        pids = self.get_job_process_list()
        timing_info = [self.get_process_times(pid) for pid in pids if pid != self.current_process]
        child_timing_info = [0, 0, 0, 0]
        for info in timing_info:
            child_timing_info[0] += info[0]
            child_timing_info[1] += info[1]
            child_timing_info[2] += info[2]
            child_timing_info[3] += info[3]
        return tuple(child_timing_info)

    def get_resource_usage(self):
        """
        Gets resource usage for both this process and all child processes in this job.

        :return: a tuple with the following values:
            - parent physical memory usage
            - parent virtual memory usage
            - parent kernel CPU time
            - parent user CPU time
            - child physical memory usage
            - child virtual memory usage
            - child kernel CPU time
            - child user CPU time
        """

        process_creation, process_exit, process_kernel, process_user = self.get_process_times()
        job_info = self.get_job_accounting_info()
        child_kernel = (job_info.BasicInfo.TotalKernelTime / 10000000) - process_kernel
        child_user = (job_info.BasicInfo.TotalUserTime / 10000000) - process_user

        process_phys_mem, process_virt_mem = self.get_current_memory_usage()
        child_phys_mem, child_virt_mem = self.get_children_memory_usage()

        return process_phys_mem, process_virt_mem, process_kernel, process_user, child_phys_mem, child_virt_mem, \
            child_kernel, child_user

    @staticmethod
    def get_error(api_name=None):
        """
        Convenience function to retrieve and format the last error generated by a Win32 API call.

        :param api_name: Name of the API call that triggered the error
        :return: RuntimeError with a formatted message describing the error
        """
        if api_name is None:
            return RuntimeError('Win32 API call failed: [WinError {}] {}'.format(
                ctypes.GetLastError(), ctypes.FormatError()
            ))
        else:
            return RuntimeError('Win32 {} API call failed: [WinError {}] {}'.format(
                api_name, ctypes.GetLastError(), ctypes.FormatError()
            ))

==> azureml_train_automl-1.0.8-py3-none-any.whl: azureml/train/automl/_win32_helper.py <==
# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""_win32_helper.py, Helper code that wraps the Win32 API using ctypes."""

import ctypes
from ctypes import wintypes
from ctypes import windll


class PROCESS_MEMORY_COUNTERS_EX(ctypes.Structure):
    """Keep track of process memory counters."""

    _fields_ = [('cb', wintypes.DWORD),
                ('PageFaultCount', wintypes.DWORD),
                ('PeakWorkingSetSize', wintypes.WPARAM),
                ('WorkingSetSize', wintypes.WPARAM),
                ('QuotaPeakPagedPoolUsage', wintypes.WPARAM),
                ('QuotaPagedPoolUsage', wintypes.WPARAM),
                ('QuotaPeakNonPagedPoolUsage', wintypes.WPARAM),
                ('QuotaNonPagedPoolUsage', wintypes.WPARAM),
                ('PagefileUsage', wintypes.WPARAM),
                ('PeakPagefileUsage', wintypes.WPARAM),
                ('PrivateUsage', wintypes.WPARAM)]


class IO_COUNTERS(ctypes.Structure):
    """Keep track of process input output counters."""

    _fields_ = [('ReadOperationCount', wintypes.ULARGE_INTEGER),
                ('WriteOperationCount', wintypes.ULARGE_INTEGER),
                ('OtherOperationCount', wintypes.ULARGE_INTEGER),
                ('ReadTransferCount', wintypes.ULARGE_INTEGER),
                ('WriteTransferCount', wintypes.ULARGE_INTEGER),
                ('OtherTransferCount', wintypes.ULARGE_INTEGER)]


class JOBOBJECT_BASIC_ACCOUNTING_INFORMATION(ctypes.Structure):
    """Keep track of job accounting counters."""

    _fields_ = [('TotalUserTime', wintypes.LARGE_INTEGER),
                ('TotalKernelTime', wintypes.LARGE_INTEGER),
                ('ThisPeriodTotalUserTime', wintypes.LARGE_INTEGER),
                ('ThisPeriodTotalKernelTime', wintypes.LARGE_INTEGER),
                ('TotalPageFaultCount', wintypes.DWORD),
                ('TotalProcesses', wintypes.DWORD),
                ('ActiveProcesses', wintypes.DWORD),
                ('TotalTerminatedProcesses', wintypes.DWORD)]


class JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION(ctypes.Structure):
    """Keep track of job accounting and input output counters."""

    _fields_ = [('BasicInfo', JOBOBJECT_BASIC_ACCOUNTING_INFORMATION),
                ('IoInfo', IO_COUNTERS)]


class JobObjectInfoClass(object):
    """Keep information for job accounting."""

    JobObjectBasicAccountingInformation = 1
    JobObjectBasicAndIoAccountingInformation = 8
    JobObjectBasicLimitInformation = 2
    JobObjectBasicProcessIdList = 3
    JobObjectBasicUIRestrictions = 4
    JobObjectCpuRateControlInformation = 15
    JobObjectEndOfJobTimeInformation = 6
    JobObjectExtendedLimitInformation = 9
    JobObjectGroupInformation = 11
    JobObjectGroupInformationEx = 14
    JobObjectLimitViolationInformation = 13
    JobObjectLimitViolationInformation2 = 35
    JobObjectNetRateControlInformation = 32
    JobObjectNotificationLimitInformation = 12
    JobObjectNotificationLimitInformation2 = 34


class Win32Helper(object):
    """Handle all process interactions for a windows process."""

    def __init__(self):
        """Get a handle to ourselves for future use."""
        # HANDLE GetCurrentProcess()
        self.current_process = windll.kernel32.GetCurrentProcess()

        self.child_handles = {}

        if not self.is_process_in_job():
            # Create a job with the name automl-{PID} and assign ourselves to it. All child processes will inherit
            # this job object so we can look them up later.
            # We don't need to hold a reference to the job object because passing NULL instead of a job handle
            # will make the Win32 API use this process's job.

            # HANDLE CreateJobObjectW(LPSECURITY_ATTRIBUTES lpJobAttributes, LPWSTR lpName)
            job_object = windll.kernel32.CreateJobObjectW(None, 'automl-{}'.format(self.get_process_id()))

            self.assign_process_to_job_object(job_object)

    def __del__(self):
        """Clean up all handles to child processes to prevent memory leaks."""
        for pid in self.child_handles:
            self.close_handle(self.child_handles[pid])

    def close_handle(self, handle):
        """
        Wrap the Win32 CloseHandle function.

        BOOL WINAPI CloseHandle(
          _In_ HANDLE hObject
        );

        :param handle: A valid handle to an open object.
        """
        success = windll.kernel32.CloseHandle(handle)
        if success == 0:
            raise self.get_error('CloseHandle')

    def get_process_id(self):
        """
        Wrap the Win32 GetProcessId function.

        DWORD GetProcessId(
          HANDLE Process
        );

        :return: the pid of the current process
        """
        return windll.kernel32.GetProcessId(self.current_process)

    def open_process(self, pid=None):
        """
        Wrap the Win32 OpenProcess function.

        Note: If a handle was already opened, the previously opened handle is returned instead. This is to prevent
        memory leaks from opening too many handles and then failing to close them. This way, we only need to close
        handles once (when this object is destroyed).
        HANDLE OpenProcess(
          DWORD dwDesiredAccess,
          BOOL  bInheritHandle,
          DWORD dwProcessId
        );

        :param pid: The PID of the process to create a handle for. If None, uses the current process.
        :return: the handle of the process
        """
        if pid is None:
            return self.current_process

        if pid in self.child_handles:
            return self.child_handles[pid]

        PROCESS_ALL_ACCESS = (0x000F0000 | 0x00100000 | 0xFFF)
        handle = windll.kernel32.OpenProcess(PROCESS_ALL_ACCESS, False, pid)
        if handle == 0:
            raise self.get_error('OpenProcess')
        self.child_handles[pid] = handle
        return handle

    def get_process_memory_info(self, pid=None):
        """
        Wrap the Win32 GetProcessMemoryInfo function.

        BOOL GetProcessMemoryInfo(
          HANDLE                   Process,
          PPROCESS_MEMORY_COUNTERS ppsmemCounters,
          DWORD                    cb
        );

        :param pid: ID of the process we want to get memory usage for. If None, gets usage for current process.
        :return: a PROCESS_MEMORY_COUNTERS_EX struct
        """
        handle = self.open_process(pid)
        pmc = PROCESS_MEMORY_COUNTERS_EX()
        success = windll.psapi.GetProcessMemoryInfo(handle, ctypes.byref(pmc), ctypes.sizeof(pmc))
        if success == 0:
            raise self.get_error('GetProcessMemoryInfo')
        return pmc

    def assign_process_to_job_object(self, job, pid=None):
        """
        Wrap the Win32 AssignProcessToJobObject function.

        BOOL WINAPI AssignProcessToJobObject(
          _In_ HANDLE hJob,
          _In_ HANDLE hProcess
        );
        :param job: Handle to the job object we want to assign to.
        :param pid: ID of the process we want to assign to the job. If None, assigns the current process to the job.
        """
        handle = self.open_process(pid)
        success = windll.kernel32.AssignProcessToJobObject(job, handle)
        if success == 0:
            raise self.get_error('AssignProcessToJobObject')

    def is_process_in_job(self):
        """
        Wrap the Win32 IsProcessInJob function.

        BOOL WINAPI IsProcessInJob(
          _In_     HANDLE ProcessHandle,
          _In_opt_ HANDLE JobHandle,
          _Out_    PBOOL  Result
        );

        :return: true if this process is running inside a job, false otherwise
        """
        result = ctypes.c_bool()
        success = windll.kernel32.IsProcessInJob(self.current_process, None, ctypes.byref(result))

        if success == 0:
            raise Win32Helper.get_error('IsProcessInJob')

        return result.value

    def query_information_job_object(self, job_object_info_class, job_info_struct):
        """
        Wrap the Win32 QueryInformationJobObject function.

        Queries info for the job this process is in.
        BOOL WINAPI QueryInformationJobObject(
          _In_opt_  HANDLE             hJob,
          _In_      JOBOBJECTINFOCLASS JobObjectInfoClass,
          _Out_     LPVOID             lpJobObjectInfo,
          _In_      DWORD              cbJobObjectInfoLength,
          _Out_opt_ LPDWORD            lpReturnLength
        );

        :param job_object_info_class: the info class that should be queried
        :param job_info_struct: a struct to pass to the query
        :return: a struct with the job info, type depends on which info class is specified
        """
        out_size = wintypes.DWORD()

        success = windll.kernel32.QueryInformationJobObject(None,
                                                            job_object_info_class,
                                                            ctypes.byref(job_info_struct),
                                                            ctypes.sizeof(job_info_struct),
                                                            ctypes.byref(out_size))
        if success == 0:
            raise self.get_error('QueryInformationJobObject')
        return job_info_struct

    def get_process_times(self, pid=None):
        """
        Wrap the Win32 GetProcessTimes function.

        Retrieves timing information for the specified process.
        Note: The Win32 API returns times in 100ns increments, but this function converts to seconds for convenience.
        BOOL GetProcessTimes(
          HANDLE     hProcess,
          LPFILETIME lpCreationTime,
          LPFILETIME lpExitTime,
          LPFILETIME lpKernelTime,
          LPFILETIME lpUserTime
        );

        :param pid: The PID of the process to get timing information for. If None, gets info for the current process.
        :return: a tuple of ints containing (creation time, exit time, kernel time, user time) in seconds
        """
        handle = self.open_process(pid)

        creation_time = wintypes.FILETIME()
        exit_time = wintypes.FILETIME()
        kernel_time = wintypes.FILETIME()
        user_time = wintypes.FILETIME()

        success = windll.kernel32.GetProcessTimes(handle,
                                                  ctypes.byref(creation_time),
                                                  ctypes.byref(exit_time),
                                                  ctypes.byref(kernel_time),
                                                  ctypes.byref(user_time))
        if success == 0:
            raise self.get_error('GetProcessTimes')

        def ft_to_sec(ft):
            total = ft.dwHighDateTime << 32
            total += ft.dwLowDateTime
            return total / 10000000

        return ft_to_sec(creation_time), ft_to_sec(exit_time), ft_to_sec(kernel_time), ft_to_sec(user_time)

    def get_current_memory_usage(self):
        """
        Get the physical and virtual memory usage of this process.

        :return: a tuple containing [physical memory used, virtual memory used] in bytes
        """
        mem_info = self.get_process_memory_info()
        return mem_info.WorkingSetSize, mem_info.PrivateUsage

    def get_children_memory_usage(self):
        """
        Get the physical and virtual memory usage of all of this process's children.

        :return: a tuple containing [physical memory used, virtual memory used] in bytes
        """
        pids = self.get_job_process_list()
        mem_info = [self.get_process_memory_info(pid) for pid in pids if pid != self.current_process]
        child_mem_usage = [0, 0]
        for info in mem_info:
            child_mem_usage[0] += info.WorkingSetSize
            child_mem_usage[1] += info.PrivateUsage
        return tuple(child_mem_usage)

    def get_job_accounting_info(self):
        """
        Get the basic accounting and I/O accounting information for the job this process is in.

        :return: a JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION struct
        """
        struct = JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION()
        return self.query_information_job_object(JobObjectInfoClass.JobObjectBasicAndIoAccountingInformation, struct)

    def get_job_process_list(self):
        """
        Get the list of pids of all processes associated with the job this process is in.

        :return: a list of all pids
        """
        captured_all_processes = False
        max_processes = 10
        struct = None
        while not captured_all_processes:
            # This must be defined dynamically because the struct is variable length
            class JobObjectBasicProcessIdListStruct(ctypes.Structure):
                _fields_ = [('NumberOfAssignedProcesses', wintypes.DWORD),
                            ('NumberOfProcessIdsInList', wintypes.DWORD),
                            ('ProcessIdList', wintypes.WPARAM * max_processes)]
            struct = JobObjectBasicProcessIdListStruct()
            struct.NumberOfAssignedProcesses = max_processes
            self.query_information_job_object(JobObjectInfoClass.JobObjectBasicProcessIdList, struct)
            if struct.NumberOfProcessIdsInList > struct.NumberOfAssignedProcesses:
                # Not enough space to capture all pids, resize array and try again
                max_processes *= 2
            else:
                captured_all_processes = True
        return list(struct.ProcessIdList[:struct.NumberOfProcessIdsInList])

    def get_child_process_times(self):
        """
        Get the timing information for all currently running processes in this job, except this one.

        :return: a tuple of ints containing (creation time, exit time, kernel time, user time) in seconds
        """
        pids = self.get_job_process_list()
        timing_info = [self.get_process_times(pid) for pid in pids if pid != self.current_process]
        child_timing_info = [0, 0, 0, 0]
        for info in timing_info:
            child_timing_info[0] += info[0]
            child_timing_info[1] += info[1]
            child_timing_info[2] += info[2]
            child_timing_info[3] += info[3]
        return tuple(child_timing_info)

    def get_resource_usage(self):
        """
        Get the resource usage for both this process and all child processes in this job.

        :return: a tuple with the following values:
            - parent physical memory usage
            - parent virtual memory usage
            - parent kernel CPU time
            - parent user CPU time
            - child physical memory usage
            - child virtual memory usage
            - child kernel CPU time
            - child user CPU time
        """
        process_creation, process_exit, process_kernel, process_user = self.get_process_times()
        job_info = self.get_job_accounting_info()
        child_kernel = (job_info.BasicInfo.TotalKernelTime / 10000000) - process_kernel
        child_user = (job_info.BasicInfo.TotalUserTime / 10000000) - process_user

        process_phys_mem, process_virt_mem = self.get_current_memory_usage()
        child_phys_mem, child_virt_mem = self.get_children_memory_usage()

        return process_phys_mem, process_virt_mem, process_kernel, process_user, child_phys_mem, child_virt_mem, \
            child_kernel, child_user

    @staticmethod
    def get_error(api_name=None):
        """
        Retrieve and format the last error generated by a Win32 API call.

        :param api_name: Name of the API call that triggered the error
        :return: RuntimeError with a formatted message describing the error
        """
        if api_name is None:
            return RuntimeError('Win32 API call failed: [WinError {}] {}'.format(
                ctypes.GetLastError(), ctypes.FormatError()
            ))
        else:
            return RuntimeError('Win32 {} API call failed: [WinError {}] {}'.format(
                api_name, ctypes.GetLastError(), ctypes.FormatError()
            ))
